#include <stdio.h>
#include <stdlib.h>
#include "mex.h"
#include "matrix.h"
#include <math.h>
#include <cuda_runtime.h>

#define THREADS_PER_BLOCK	512
#define VERBOSE 0

/*************************
 * bvToolbox  Copyright (C) 2013  PIERRE Fabien (read license.txt)
 * Auteur 	: Fabien PIERRE.
 * Version	:	1.1.
 * Date	    : 30 mai 2013.
*/
 
/*Kernel gradient*/
__global__ void
	gradient( const float *image_initiale, float *gradient_x, float *gradient_y,int height, int width)
{
	int i = blockDim.x * blockIdx.x + threadIdx.x;

	int lin= i % height;
	int col= (i-lin)/height;

	if(lin==(height-1))
	{
		gradient_x[lin+col*height]=0;
	}
	else
	{
		gradient_x[lin+col*height]=
			image_initiale[lin+(col)*height+1]
		-image_initiale[lin+col*height];
	}
	if((col==(width-1)))
	{
		gradient_y[lin+col*height]= 0;
	}
	else
	{
		gradient_y[lin+col*height]=
			image_initiale[lin+(col+1)*height]
		-image_initiale[lin+col*height];
	}
}

/*Kernel divergence*/
__global__ void
	divergence( const float *gradient_x, const float *gradient_y, float *result, int height, int width)
{
	int i = blockDim.x * blockIdx.x + threadIdx.x;

	int lin= i % height;
	int col= (i-lin)/height;

	float x_1;
	float x_2;
	if(lin==0)
	{
		x_1=gradient_x[lin+height*col];
	}
	else if (lin==(height-1))
	{
		x_1=-gradient_x[lin+height*col-1];
	}
	else
	{
		x_1=gradient_x[lin+height*col]
		-gradient_x[lin+height*col-1];
	}

	if (col==0)
	{
		x_2=gradient_y[lin+height*col];
	}
	else if (col==(width-1))
	{
		x_2=-gradient_y[lin+height*(col-1)];
	}
	else
	{
		x_2=gradient_y[lin+height*col]
		-gradient_y[lin+height*(col-1)];
	}
	result[lin+height*col]= x_1+x_2;
}

/*Kernel de calcul intermédiaire : 'fonction1'.*/
__global__ void
	fonction1( float *G, const float *dv0, const float *g, const float lambda, int height, int width)
{	
	int i = blockDim.x * blockIdx.x + threadIdx.x;
	if(i<height*width){
		G[i] = dv0[i] - g[i] / lambda;
	}
}

/*Kernel de calcul intermédiaire : 'fonction2'.*/
__global__ void
	fonction2( float *Im1, float *Im2, float *G, const float *G1, const float *G2, int height, int width)
{	
	int i = blockDim.x * blockIdx.x + threadIdx.x;
	if(i<height*width){

		G[i]=sqrt(G1[i] * G1[i] + G2[i] * G2[i]);
		Im2[i] = (Im2[i] + 0.2f * G2[i]) / (1.0f + 0.2f * G[i]);
		Im1[i] = (Im1[i] + 0.2f * G1[i]) / (1.0f + 0.2f * G[i]);
	}
}

/*Kernel de calcul intermédiaire : 'fonction3'.*/
__global__ void
	fonction3(   const float *g,  float *Im, const float lambda,  int height, int width)
{
	int i = blockDim.x * blockIdx.x + threadIdx.x;
	if(i<height*width){
		Im[i] = g[i] - lambda * Im[i];
	}
}

void bvIteratif( double* result, double* Im_init,  int itt, double lambda, int height, int width){
	if(Im_init != NULL){

		float *g_host, *dv0_host;
		int N= height*width;
		size_t size= N*sizeof(float);
		g_host= (float*)malloc(size);
		dv0_host= (float*)malloc(size);
		float lambda_f= (float)lambda;

		/*Conversion en float.*/
		for(int i=0; i<N; i++){
			g_host[i]= (float)Im_init[i];
		}

		// Allocation de la variable device g_device
		float *g_device = NULL;
		cudaError_t err = cudaMalloc((void **)&g_device, size);/*check*/{
			if (err != cudaSuccess){
				mexPrintf( "Failed to allocate device vector g_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		// Allocation de la variable device G_device
		float *G_device = NULL;
		err = cudaMalloc((void **)&G_device, size);/*check*/{
			if (err != cudaSuccess){
				mexPrintf( "Failed to allocate device vector G_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		// Allocation de la variable device G1_device
		float *G1_device = NULL;
		err = cudaMalloc((void **)&G1_device, size);/*check*/{
			if (err != cudaSuccess){
				mexPrintf( "Failed to allocate device vector G1_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		// Allocation de la variable device G2_device
		float *G2_device = NULL;
		err = cudaMalloc((void **)&G2_device, size);/*check*/{
			if (err != cudaSuccess){
				mexPrintf( "Failed to allocate device vector G2_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		// Allocation de la variable device Im1_device
		float *Im1_device = NULL;
		err = cudaMalloc((void **)&Im1_device, size);/*check*/{
			if (err != cudaSuccess){
				mexPrintf( "Failed to allocate device vector Im1_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		// Allocation de la variable device Im2_device
		float *Im2_device = NULL;
		err = cudaMalloc((void **)&Im2_device, size);/*check*/{
			if (err != cudaSuccess){
				mexPrintf( "Failed to allocate device vector Im2_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		// Allocation de la variable device dv0_device
		float *dv0_device = NULL;
		err = cudaMalloc((void **)&dv0_device, size);/*check*/{
			if (err != cudaSuccess){
				mexPrintf( "Failed to allocate device vector dv0_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}

		clock_t instant1= clock();
		
		// Copy the host input vectors A and B in host memory to the device input vectors in
		// device memory
		err = cudaMemcpy(g_device, g_host, size, cudaMemcpyHostToDevice);/*check*/{
			if (err != cudaSuccess){
				mexPrintf( "Failed to copy vector droit from host to device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}

		err = cudaMemcpy(Im1_device, g_host, size, cudaMemcpyHostToDevice);/*check*/{
			if (err != cudaSuccess){
				mexPrintf( "Failed to copy vector droit from host to device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}

		err = cudaMemcpy(Im2_device, g_host, size, cudaMemcpyHostToDevice);/*check*/{
			if (err != cudaSuccess){
				mexPrintf( "Failed to copy vector droit from host to device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}

		int numElements= height*width;
		int blocksPerGrid = numElements  / THREADS_PER_BLOCK ;
			
		for(int iter=0; iter< itt; iter++){
			divergence<<<blocksPerGrid, THREADS_PER_BLOCK>>>(Im1_device, Im2_device, dv0_device, height, width);
			err = cudaGetLastError();/*check*/{
				if (err != cudaSuccess){
					mexPrintf( "Failed to launch diavergence kernel (error code %s)!\n", cudaGetErrorString(err));
					return ;
				}
			}
			fonction1<<<blocksPerGrid, THREADS_PER_BLOCK>>>(G_device,  dv0_device, g_device, lambda_f, height, width);
			err = cudaGetLastError();/*check*/{
				if (err != cudaSuccess){
					mexPrintf( "Failed to launch fonction1 kernel (error code %s)!\n", cudaGetErrorString(err));
					return ;
				}
			}
			gradient<<<blocksPerGrid, THREADS_PER_BLOCK>>>(G_device, G1_device, G2_device, height, width);
			err = cudaGetLastError();/*check*/{
				if (err != cudaSuccess){
					mexPrintf( "Failed to launch gradient kernel (error code %s)!\n", cudaGetErrorString(err));
					return ;
				}
			}
			fonction2<<<blocksPerGrid, THREADS_PER_BLOCK>>>(Im1_device, Im2_device, G_device, G1_device, G2_device, height, width);
			err = cudaGetLastError();/*check*/{
				if (err != cudaSuccess){
					mexPrintf( "Failed to launch fonction2 kernel (error code %s)!\n", cudaGetErrorString(err));
					return ;
				}
			}
		}

		divergence<<<blocksPerGrid, THREADS_PER_BLOCK>>>(Im1_device, Im2_device, dv0_device, height, width);
		err = cudaGetLastError();/*check*/{
			if (err != cudaSuccess){
				mexPrintf( "Failed to launch divergence kernel (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		fonction3<<<blocksPerGrid, THREADS_PER_BLOCK>>>(g_device, dv0_device, lambda_f,  height, width);
		err = cudaGetLastError();/*check*/{
			if (err != cudaSuccess){
				mexPrintf( "Failed to launch fonction3 kernel (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		
		err = cudaMemcpy(dv0_host, dv0_device, size, cudaMemcpyDeviceToHost);/*check*/{

		if (err != cudaSuccess)
			{
				mexPrintf( "Failed to copy vector dv0 from device to host (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		
		if(VERBOSE){
			clock_t instant2= clock();
			mexPrintf("temps GPU = %f ms.\n", difftime(instant2, instant1));
		}
		
		/*Re converion en double.*/
		for(int i=0;  i<height*width; i++){
			result[i]= (double)dv0_host[i];
		}


		err = cudaFree(g_device);/*check*/{
			if (err != cudaSuccess)
			{
				mexPrintf( "Failed to free device vector g_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		err = cudaFree(G_device);/*check*/{
			if (err != cudaSuccess)
			{
				mexPrintf( "Failed to free device vector G_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		err = cudaFree(G1_device);/*check*/{
			if (err != cudaSuccess)
			{
				mexPrintf( "Failed to free device vector G1_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		err = cudaFree(G2_device);/*check*/{
			if (err != cudaSuccess)
			{
				mexPrintf( "Failed to free device vector G2_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		err = cudaFree(Im1_device);/*check*/{
			if (err != cudaSuccess)
			{
				mexPrintf( "Failed to free device vector Im1_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		err = cudaFree(Im2_device);/*check*/{
			if (err != cudaSuccess)
			{
				mexPrintf( "Failed to free device vector Im2_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}
		err = cudaFree(dv0_device);/*check*/{
			if (err != cudaSuccess)


			{
				mexPrintf( "Failed to free device vector dv0_device (error code %s)!\n", cudaGetErrorString(err));
				return ;
			}
		}

		// Reset the device and exit
		err = cudaDeviceReset();/*check*/{
			if (err != cudaSuccess){
				mexPrintf("Failed to deinitialize the device! error=%s\n", cudaGetErrorString(err));
				return ;
			}
		}

		/*Liberation de la RAM CPU (host).*/
		free(dv0_host);
		free(g_host);

	}
	else{
		return;
	}
}


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	/* Declarations.*/
	double *result, *g;
	int height, width, ndim, N;
	const int *dims;
	double lambda;

	/* Verification du bon nombre d'argument.*/
	if(nrhs!=3) {
		mexErrMsgIdAndTxt("bvToolbox:bvProximalOperator:nrhs","3 inputs required.");
	}
	if(nlhs!=1) {
		mexErrMsgIdAndTxt("bvToolbox:bvProximalOperator:nlhs","One output required.");
	}

	/*gestion entrees.*/
	ndim= mxGetNumberOfDimensions(prhs[0]);
	dims= (const int *)(mxGetDimensions(prhs[0]));
	height = dims[0];
	width = dims[1];

	N = (int)(*(mxGetPr(prhs[1])));
	lambda = (double)(*(mxGetPr(prhs[2])));
	g = (double*)mxGetPr(prhs[0]);

	/*gestion sortie.*/
	plhs[0]= mxCreateNumericArray(ndim, (const mwSize *)dims, mxDOUBLE_CLASS, mxREAL);
	result= (double*)mxGetPr(plhs[0]);

	/*Lancement de l'algorithme de calcul paralelle.*/
	bvIteratif(result, g, N, lambda, height, width);

	return;
}
